@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#ident "$Id: vtexft.c,v 1.17 2003/05/30 08:55:28 padraigo Exp $"
+#ident "$Id: vtexft.c,v 1.24 2006/02/25 23:20:42 behdad Exp $"
 
 #include "../config.h"
 
@@ -35,6 +35,7 @@
 #include "vtedraw.h"
 #include "vtefc.h"
 #include "vtexft.h"
+#include "vtetree.h"
 
 #ifdef ENABLE_NLS
 #include <libintl.h>
@@ -53,8 +54,8 @@ struct _vte_xft_font {
 #endif
 	GArray *patterns;
 	GArray *fonts;
-	GTree *fontmap;
-	GTree *widths;
+	VteTree *fontmap;
+	VteTree *widths;
 };
 
 struct _vte_xft_data
@@ -105,14 +106,15 @@ _vte_xft_text_extents(struct _vte_xft_font *font, XftFont *ftfont, FcChar32 c,
 }
 
 static struct _vte_xft_font *
-_vte_xft_font_open(GtkWidget *widget, const PangoFontDescription *fontdesc)
+_vte_xft_font_open(GtkWidget *widget, const PangoFontDescription *fontdesc,
+		   VteTerminalAntiAlias antialias)
 {
 	struct _vte_xft_font *font;
 	GArray *patterns;
 
 	patterns = g_array_new(TRUE, TRUE, sizeof(FcPattern*));
-	if (!_vte_fc_patterns_from_pango_font_desc(widget, fontdesc, patterns,
-						   NULL, NULL)) {
+	if (!_vte_fc_patterns_from_pango_font_desc(widget, fontdesc, antialias,
+						   patterns, NULL, NULL)) {
 		g_array_free(patterns, TRUE);
 		return NULL;
 	}
@@ -123,8 +125,8 @@ _vte_xft_font_open(GtkWidget *widget, const PangoFontDescription *fontdesc)
 #endif
 	font->patterns = patterns;
 	font->fonts = g_array_new(TRUE, TRUE, sizeof(XftFont*));
-	font->fontmap = g_tree_new(_vte_xft_direct_compare);
-	font->widths = g_tree_new(_vte_xft_direct_compare);
+	font->fontmap = _vte_tree_new(_vte_xft_direct_compare);
+	font->widths = _vte_tree_new(_vte_xft_direct_compare);
 
 	return font;
 }
@@ -138,11 +140,11 @@ _vte_xft_font_close(struct _vte_xft_font *font)
 	FcPattern *pattern;
 	int i;
 
-	g_return_if_fail(font != NULL);
-	g_return_if_fail(font->patterns != NULL);
-	g_return_if_fail(font->fonts != NULL);
-	g_return_if_fail(font->fontmap != NULL);
-	g_return_if_fail(font->widths != NULL);
+	g_assert(font != NULL);
+	g_assert(font->patterns != NULL);
+	g_assert(font->fonts != NULL);
+	g_assert(font->fontmap != NULL);
+	g_assert(font->widths != NULL);
 
 	for (i = 0; i < font->patterns->len; i++) {
 		pattern = g_array_index(font->patterns, FcPattern*, i);
@@ -165,9 +167,9 @@ _vte_xft_font_close(struct _vte_xft_font *font)
 	g_array_free(font->fonts, TRUE);
 	font->fonts = NULL;
 
-	g_tree_destroy(font->fontmap);
+	_vte_tree_destroy(font->fontmap);
 	font->fontmap = NULL;
-	g_tree_destroy(font->widths);
+	_vte_tree_destroy(font->widths);
 	font->widths = NULL;
 
 	g_free(font);
@@ -183,14 +185,14 @@ _vte_xft_font_for_char(struct _vte_xft_font *font, gunichar c)
 	Display *display;
 	gpointer p = GINT_TO_POINTER(c);
 
-	g_return_val_if_fail(font != NULL, NULL);
-	g_return_val_if_fail(font->patterns != NULL, NULL);
-	g_return_val_if_fail(font->fonts != NULL, NULL);
-	g_return_val_if_fail(font->fontmap != NULL, NULL);
-	g_return_val_if_fail(font->widths != NULL, NULL);
+	g_assert(font != NULL);
+	g_assert(font->patterns != NULL);
+	g_assert(font->fonts != NULL);
+	g_assert(font->fontmap != NULL);
+	g_assert(font->widths != NULL);
 
 	/* Check if we have a char-to-font entry for it. */
-	i = GPOINTER_TO_INT(g_tree_lookup(font->fontmap, p));
+	i = GPOINTER_TO_INT(_vte_tree_lookup(font->fontmap, p));
 	if (i != 0) {
 		switch (i) {
 		/* Checked before, no luck. */
@@ -224,7 +226,7 @@ _vte_xft_font_for_char(struct _vte_xft_font *font, gunichar c)
 
 	/* Match? */
 	if (i < font->fonts->len) {
-		g_tree_insert(font->fontmap,
+		_vte_tree_insert(font->fontmap,
 			      p,
 			      GINT_TO_POINTER(i + FONT_INDEX_FUDGE));
 		ftfont = g_array_index(font->fonts, XftFont *, i);
@@ -252,7 +254,7 @@ _vte_xft_font_for_char(struct _vte_xft_font *font, gunichar c)
 
 	/* No match? */
 	if (i >= font->patterns->len) {
-		g_tree_insert(font->fontmap,
+		_vte_tree_insert(font->fontmap,
 			      p,
 			      GINT_TO_POINTER(-FONT_INDEX_FUDGE));
 		if (font->fonts->len > 0) {
@@ -280,14 +282,14 @@ _vte_xft_char_width(struct _vte_xft_font *font, XftFont *ftfont, gunichar c)
 	gpointer p = GINT_TO_POINTER(c);
 	int i;
 
-	g_return_val_if_fail(font != NULL, 0);
-	g_return_val_if_fail(font->patterns != NULL, 0);
-	g_return_val_if_fail(font->fonts != NULL, 0);
-	g_return_val_if_fail(font->fontmap != NULL, 0);
-	g_return_val_if_fail(font->widths != NULL, 0);
+	g_assert(font != NULL);
+	g_assert(font->patterns != NULL);
+	g_assert(font->fonts != NULL);
+	g_assert(font->fontmap != NULL);
+	g_assert(font->widths != NULL);
 
 	/* Check if we have a char-to-width entry for it. */
-	i = GPOINTER_TO_INT(g_tree_lookup(font->widths, p));
+	i = GPOINTER_TO_INT(_vte_tree_lookup(font->widths, p));
 	if (i != 0) {
 		switch (i) {
 		case -CHAR_WIDTH_FUDGE:
@@ -305,7 +307,7 @@ _vte_xft_char_width(struct _vte_xft_font *font, XftFont *ftfont, gunichar c)
 		_vte_xft_text_extents(font, ftfont, c, &extents);
 	}
 	i = extents.xOff + CHAR_WIDTH_FUDGE;
-	g_tree_insert(font->widths, p, GINT_TO_POINTER(i));
+	_vte_tree_insert(font->widths, p, GINT_TO_POINTER(i));
 	return extents.xOff;
 }
 
@@ -530,7 +532,8 @@ _vte_xft_clear(struct _vte_draw *draw,
 
 static void
 _vte_xft_set_text_font(struct _vte_draw *draw,
-		       const PangoFontDescription *fontdesc)
+		       const PangoFontDescription *fontdesc,
+		       VteTerminalAntiAlias antialias)
 {
 	GString *string;
 	XftFont *font;
@@ -546,7 +549,7 @@ _vte_xft_set_text_font(struct _vte_draw *draw,
 		_vte_xft_font_close(data->font);
 		data->font = NULL;
 	}
-	data->font = _vte_xft_font_open(draw->widget, fontdesc);
+	data->font = _vte_xft_font_open(draw->widget, fontdesc, antialias);
 
 	draw->width = 1;
 	draw->height = 1;
@@ -622,6 +625,20 @@ _vte_xft_get_text_ascent(struct _vte_draw *draw)
 	return draw->ascent;
 }
 
+static int
+_vte_xft_get_char_width(struct _vte_draw *draw, gunichar c, int columns)
+{
+	struct _vte_xft_data *data;
+	XftFont *ftfont;
+
+	data = (struct _vte_xft_data*) draw->impl_data;
+	ftfont = _vte_xft_font_for_char(data->font, c);
+	if (ftfont == NULL) {
+		return _vte_xft_get_text_width(draw) * columns;
+	}
+	return _vte_xft_char_width(data->font, ftfont, c);
+}
+
 static gboolean
 _vte_xft_get_using_fontconfig(struct _vte_draw *draw)
 {
@@ -645,6 +662,7 @@ _vte_xft_drawcharfontspec(XftDraw *draw, XftColor *color,
 			  XftCharFontSpec *specs, int n)
 {
 	int i, j;
+
 	i = j = 0;
 	while (i < n) {
 		for (j = i + 1; j < n; j++) {
@@ -679,7 +697,7 @@ _vte_xft_draw_text(struct _vte_draw *draw,
 	for (i = j = 0; i < n_requests; i++) {
 		specs[j].font = _vte_xft_font_for_char(data->font,
 						       requests[i].c);
-		if (specs[j].font != NULL) {
+		if (specs[j].font != NULL && requests[i].c != 32) {
 			specs[j].x = requests[i].x - data->x_offs;
 			width = _vte_xft_char_width(data->font,
 						    specs[j].font,
@@ -692,7 +710,7 @@ _vte_xft_draw_text(struct _vte_draw *draw,
 			specs[j].y = requests[i].y - data->y_offs + draw->ascent;
 			specs[j].ucs4 = requests[i].c;
 			j++;
-		} else {
+		} else if (requests[i].c != 32) {
 			g_warning(_("Can not draw character U+%04x.\n"),
 				  requests[i].c);
 		}
@@ -816,6 +834,7 @@ struct _vte_draw_impl _vte_draw_xft = {
 	_vte_xft_get_text_width,
 	_vte_xft_get_text_height,
 	_vte_xft_get_text_ascent,
+	_vte_xft_get_char_width,
 	_vte_xft_get_using_fontconfig,
 	_vte_xft_draw_text,
 	_vte_xft_draw_char,
